# basically code from the victronenergy community pages, consolidated:
# sdm630v2.zip from telekatz (SDM630v2) post Nov 10 2023
# Eastron.zip from sega (SDM72) post Dec 05 2023
# SDM120M.zip from robertik (SDM120M) post  Dec 09 2023
# Note: merged and tested with SDM630 MODBUS V3 by woz001, Dec 31 2023
# Checked against this openly available documents found by a startpage search:
# https://www.eastroneurope.com/images/uploads/products/protocol/SDM630_MODBUS_Protocol.pdf
# https://www.eastroneurope.com/images/uploads/products/protocol/1_SDM72CT-M_Protocol_V1.1.pdf
# https://www.eastroneurope.com/images/uploads/products/protocol/SDM120-MODBUS_Protocol.pdf
# --> assuming this are standard Chinese DL/T 645-2007 devices...
import logging

import eastron_device as device
import probe
from register import *

log = logging.getLogger()

# generic, would be nicer if this could be added to the register lib...
class Reg_f32b(Reg_num):
    coding = ('>f', '>2H')
    count = 2
    rtype = float

# ----------------------------------------------------------------------------

nr_phases = [ 0, 1, 3, 3 ]

phase_configs = [
    'undefined',
    '1p2w',
    '3p3w',
    '3p4w',
]

# ----------------------------------------------------------------------------

class Eastron_Meter_1P(device.CustomName, device.EnergyMeter):
    productid = 0xb023 # id assigned by Victron Support
    productname = 'Eastron Energy Meter'
    min_timeout = 0.5

    def device_init(self):
        self.info_regs = [
            Reg_u16( 0xfc02, '/HardwareVersion'),
            Reg_u16( 0xfc03, '/FirmwareVersion'),
            Reg_u32b(0x0014, '/Serial'),
        ]

        self.read_info()

        regs = [
            Reg_f32b(0x0034, '/Ac/Power',          1, '%.1f W'),
            Reg_f32b(0x0030, '/Ac/Current',        1, '%.1f A'),
            Reg_f32b(0x0046, '/Ac/Frequency',      1, '%.1f Hz'),
            Reg_f32b(0x0048, '/Ac/Energy/Forward', 1, '%.1f kWh'),
            Reg_f32b(0x004a, '/Ac/Energy/Reverse', 1, '%.1f kWh'),
            Reg_f32b(0x0000, '/Ac/L1/Voltage',        1, '%.1f V'),
            Reg_f32b(0x0006, '/Ac/L1/Current',        1, '%.2f A'),
            Reg_f32b(0x000c, '/Ac/L1/Power',          1, '%.1f W'),
            Reg_f32b(0x0048, '/Ac/L1/Energy/Forward', 1, '%.1f kWh'),
            Reg_f32b(0x004a, '/Ac/L1/Energy/Reverse', 1, '%.1f kWh'),
        ]

        self.data_regs = regs
        self.phaseconfig = phase_configs[1]
        self.nr_phases = nr_phases[1]

    def dbus_write_register(self, reg, path, val):
        super().dbus_write_register(reg, path, val)
        self.sched_reinit()

    def get_ident(self):
        return 'cg_%s' % self.info['/Serial']

# ----------------------------------------------------------------------------

class Eastron_Meter_3P(Eastron_Meter_1P):
    def phase_regs(self, n):
        s = 2 * (n - 1)
        return [
            Reg_f32b(0x0000 + s, '/Ac/L%d/Voltage' % n,        1, '%.1f V'),
            Reg_f32b(0x0006 + s, '/Ac/L%d/Current' % n,        1, '%.1f A'),
            Reg_f32b(0x000c + s, '/Ac/L%d/Power' % n,          1, '%.1f W'),
            Reg_f32b(0x015a + s, '/Ac/L%d/Energy/Forward' % n, 1, '%.1f kWh'),
            Reg_f32b(0x0160 + s, '/Ac/L%d/Energy/Reverse' % n, 1, '%.1f kWh'),
        ]

    def device_init(self):
        self.info_regs = [
            Reg_u16( 0xfc02, '/HardwareVersion'),
            Reg_u16( 0xfc03, '/FirmwareVersion'),
            Reg_f32b(0x000a, '/PhaseConfig', text=phase_configs, write=(0, 3)),
            Reg_u32b(0x0014, '/Serial'),
        ]

        self.read_info()

        phases = nr_phases[int(self.info['/PhaseConfig'])]

        regs = [
            Reg_f32b(0x0034, '/Ac/Power',          1, '%.1f W'),
            Reg_f32b(0x0030, '/Ac/Current',        1, '%.1f A'),
            Reg_f32b(0x0046, '/Ac/Frequency',      1, '%.1f Hz'),
            Reg_f32b(0x0048, '/Ac/Energy/Forward', 1, '%.1f kWh'),
            Reg_f32b(0x004a, '/Ac/Energy/Reverse', 1, '%.1f kWh'),
        ]

        for n in range(1, phases + 1):
            regs += self.phase_regs(n)

        self.data_regs = regs
        self.nr_phases = phases

# ----------------------------------------------------------------------------

models = {
    32: {
        'model':    'SDM120M',
        'handler':  Eastron_Meter_1P,
    },
    112: {
        'model':    'SDM630Modbusv2',
        'handler':  Eastron_Meter_3P,
    },
    121: {
        'model':    'SDM630Modbusv2',
        'handler':  Eastron_Meter_3P,
    },
    135: {
        'model':    'SDM72Dv2',
        'handler':  Eastron_Meter_3P,
    },
    137: {
        'model':    'SDM72Modbusv2',
        'handler':  Eastron_Meter_3P,
    },
}

# ----------------------------------------------------------------------------

probe.add_handler(probe.ModelRegister(Reg_u16(0xfc02), models,
                                      methods=['rtu', 'tcp'],
                                      rates=[9600, 19200, 38400],
                                      units=[1, 2]))